#PyMT gesture recognition 
#Code has been adapted for a multi-touch gesture recognition system by Espen S. Nygaard
#Only parts of the original file remains

'''
Gesture: Base for gesture recognition.

You can easily use these class to create
new gesture, and compare them ! ::

    from pymt import *

    # Create a gesture
    g = Gesture()
    g.add_stroke(point_list=[(1,1), (3,4), (2,1)])
    g.normalize()

    # Add him to database
    gdb = GestureDatabase()
    gdb.add_gesture(g)

    # And for the next gesture, try to find him !
    g2 = Gesture()
    # ...
    gdb.find(g2)
'''

__all__ = ['Gesture', 'GestureDatabase', 'GesturePoint', 'GestureStroke']

import math
from vector import *
import anngesture

class GestureDatabase(object):
    '''Class to handle a gesture database.'''
    def __init__(self):
        self.db = []
        self.trainedAnnImage = None
        self.trainedAnnDirection = None

    def add_gesture(self, gesture):
        '''Add a new gesture in database'''
        self.db.append(gesture)

    def find(self, gesture, minscore=0.9, rotation_invariant=True, recognitionMethod = 'best'):
        '''Find current gesture in database'''
        if not gesture:
            return

        if recognitionMethod == 'best':
            return self.findBest(gesture, minscore=0.1, rotation_invariant="False", recognitionMethod = 'best')

        if recognitionMethod == 'annImage':
            score, label = gesture.get_score(gesture, rotation_invariant,recognitionMethod,self.trainedAnnImage)
            simGesture = Gesture()
            simGesture.id = label
            simGesture.label = label
            return (score, simGesture)

        if recognitionMethod == 'annDirection':
            score, label = gesture.get_score(gesture, rotation_invariant,recognitionMethod,self.trainedAnnDirection)
            simGesture = Gesture()
            simGesture.id = label
            simGesture.label = label
            return (score, simGesture)


        best = None
        bestscore = minscore
        for g in self.db:
            score = g.get_score(gesture, rotation_invariant,recognitionMethod,self.trainedAnnImage)
            if score < bestscore:
                continue
            bestscore = score
            best = g
        if not best:
            return
        return (bestscore, best)

    def gesture_to_str(self, gesture):
        '''Convert a gesture into a unique string'''
        from cStringIO import StringIO
        import pickle, base64, zlib
        io = StringIO()
        p = pickle.Pickler(io)
        p.dump(gesture)
        data = base64.b64encode(zlib.compress(io.getvalue(), 9))
        return data

    def str_to_gesture(self, data):
        '''Convert a unique string to a gesture'''
        from cStringIO import StringIO
        import pickle, base64, zlib
        io = StringIO(zlib.decompress(base64.b64decode(data)))
        p = pickle.Unpickler(io)
        gesture = p.load()
        return gesture

    def saveGestureToFile(self, filename, gesture):
        stringRep = self.gesture_to_str(gesture)
        f = open(filename,'a')
        f.write(gesture.label+','+gesture.id+','+stringRep+'\n')
        f.close()

    def loadGesturesFromFile(self, filename,addToDB=False):
        gestureList = []
        f = open(filename,'r')
        for line in f:
            gestureObj = line.split(',')
            g = self.str_to_gesture(gestureObj[2].strip())
            g.label = gestureObj[0]
            g.id = gestureObj[1]
            if addToDB == True:
                self.add_gesture(g)
            print "Loading gesture with label: "+str(g.label)+" (id="+str(g.id)+")"
            gestureList.append([gestureObj[0],gestureObj[1],g])
        f.close()
        return gestureList

    def loadAllTestSets(self,filenames):
        testSets = []
        for name in filenames:
            testSets.append(self.loadGesturesFromFile(name,addToDB=False))
        return testSets

    def trainAnnImageFromFiles(self, filenames,nrSymbols,nrPerSymbol):
        #Return a trained ANN

        #now it is 7 symbols, 5 per symbol
        allSets = self.loadAllTestSets(filenames)
        trainingPatterns = []
        for set in allSets:
            #Train nrSymbols gestures, nrPerSymbol first per symbol
            counter = 0
            for g in set:
                if  1==1: #counter%nrPerSymbol == 2 or counter%nrPerSymbol == 3:
                    patternNr = counter/nrPerSymbol #return which pattern number we are on
                    image = g[2].create2dMapFromStroke(g[2])
                    #print len(image)
                    trainingPatterns.append([image,self._generateAnnOutputPattern(nrSymbols,patternNr)])


                    print str(g[0])+"[",
                    for a in trainingPatterns[-1][0]:
                        print ("%.1f" % a),
                    print "] -> [",
                    for a in trainingPatterns[-1][1]:
                        print ("%.2f" % a),
                    print "]"
                counter = counter + 1
                
            

        #Create an ANN with 31 inputnodes, 7 hidden and nrSymbols output nodes
        n = anngesture.NN(((8*8)), 7, 7) #10 gives ok # 7 gives even better!
        #tested with this: filenames = ["testSet1275316942.37.txt", "testSet1275308118.67.txt", "testSet1275308179.27.txt", "testSet1275308229.93.txt","staale.txt","morten.txt"]
       
        # train it with some the testSet
        n.train(trainingPatterns, iterations=1000, N=0.5, M=0.1)
        self.trainedAnnImage = n
        return n


    def trainAnnDirectionFromFiles(self, filenames,nrSymbols,nrPerSymbol):
        #Return a trained ANN

        #now it is 7 symbols, 5 per symbol
        allSets = self.loadAllTestSets(filenames)
        trainingPatterns = []
        for set in allSets:
            #Train nrSymbols gestures, nrPerSymbol first per symbol
            counter = 0
            for g in set:
                if 1==1:#counter%nrPerSymbol == 2: #0 or counter%nrPerSymbol == 1:
                    patternNr = counter/nrPerSymbol #return which pattern number we are on
                    my_strokeList, dirList ,totalPoints = g[2]._calculate_direction_list(g[2])
                    for i in range(len(dirList[0])):
                        dirList[0][i] = dirList[0][i]/7.0

                    
                    trainingPatterns.append([dirList[0],self._generateAnnOutputPattern(nrSymbols,patternNr)])
                  

                    print str(g[0])+"[",
                    for a in trainingPatterns[-1][0]:
                        print ("%.1f" % a),
                    print "] -> [",
                    for a in trainingPatterns[-1][1]:
                        print ("%.2f" % a),
                    print "]"
                counter = counter + 1
            break
        #Create an ANN with 31 inputnodes, 7 hidden and nrSymbols output nodes
        n = anngesture.NN(31, 7, 7) #10 gives ok # 7 gives even better!
        #tested with this: filenames = ["testSet1275316942.37.txt", "testSet1275308118.67.txt", "testSet1275308179.27.txt", "testSet1275308229.93.txt","staale.txt","morten.txt"]

        # train it with some the testSet
        n.train(trainingPatterns)
        self.trainedAnnDirection = n
        return n

    def _generateAnnOutputPattern(self, nr, active):
        out = []

        for i in range(0,nr):
            if i == active:
                out.append(1.0)
            else:
                out.append(0.0)
        return out

    def findBest(self, gesture, minscore=0.9, rotation_invariant=True, recognitionMethod = 'dot_product'):
        '''Find current gesture in database'''
        if not gesture:
            return

        best = None
        bestscore = minscore
        for g in self.db:
            dot = g.get_score(gesture, rotation_invariant=False,recognitionMethod = 'dot_product')
            dir = g.get_score(gesture, rotation_invariant=False,recognitionMethod = 'direction_levenshtein')
            cstm = g.get_score(gesture, rotation_invariant=False,recognitionMethod = 'direction_custom')
            reg = g.get_score(gesture, rotation_invariant=False,recognitionMethod = 'region_levenshtein')
            score = dot+dir+cstm+reg
            if score < bestscore:
                continue
            bestscore = score
            best = g
        if not best:
            return
        return (bestscore, best)

class GesturePoint:
    def __init__(self, x, y):
        '''Stores the x,y coordinates of a point in the gesture'''
        self.x = float(x)
        self.y = float(y)

    def scale(self, factor):
        ''' Scales the point by the given factor '''
        self.x *= factor
        self.y *= factor
        return self

    def __repr__(self):
        return "Mouse_point:%f,%f"%(self.x,self.y)

class GestureStroke:
    ''' Gestures can be made up of multiple strokes '''
    def __init__(self):
        ''' A stroke in the gesture '''
        self.points = list()
        self.screenpoints = list()

    # These return the min and max coordinates of the stroke
    @property
    def max_x(self):
        if len(self.points) == 0:
            return 0
        return max(self.points, key = lambda pt: pt.x).x
    @property
    def min_x(self):
        if len(self.points) == 0:
            return 0
        return min(self.points, key = lambda pt: pt.x).x
    @property
    def max_y(self):
        if len(self.points) == 0:
            return 0
        return max(self.points, key = lambda pt: pt.y).y
    @property
    def min_y(self):
        if len(self.points) == 0:
            return 0
        return min(self.points, key = lambda pt: pt.y).y

    def add_point(self, x, y):
        '''
        add_point(x=x_pos, y=y_pos)
        Adds a point to the stroke
        '''
        self.points.append(GesturePoint(x, y))
        self.screenpoints.append((x, y))

    def scale_stroke(self, scale_factor):
        '''
        scale_stroke(scale_factor=float)
        Scales the stroke down by scale_factor
        '''
        self.points = map(lambda pt: pt.scale(scale_factor), self.points)

    def points_distance(self, point1, point2):
        '''
        points_distance(point1=GesturePoint, point2=GesturePoint)
        Returns the distance between two GesturePoint
        '''
        x = point1.x - point2.x
        y = point1.y - point2.y
        return math.sqrt(x*x + y*y)

    def stroke_length(self, point_list=None):
        '''
        stroke_length([point_list])
        Finds the length of the stroke. If a point list is given, finds the length of that list
        '''
        if point_list is None:
            point_list = self.points
        gesture_length = 0.0
        if len(point_list) <= 1: # If there is only one point, there is no length
            return gesture_length
        for i in xrange(len(point_list)-1):
            gesture_length += self.points_distance(point_list[i], point_list[i+1])
        return gesture_length

    def normalize_stroke(self, sample_points = 32):
        '''
        normalize_stroke([sample_points=int])
        Normalizes strokes so that every stroke has a standard number of points. Returns True if
        stroke is normalized, False if it can't be normalized. sample_points control the resolution of the stroke
        '''
        # If there is only one point or the length is 0, don't normalize
        if len(self.points) <= 1 or self.stroke_length(self.points) == 0.0:
            return False

        # Calculate how long each point should be in the stroke
        target_stroke_size = self.stroke_length(self.points)/float(sample_points)
        new_points = list()
        new_points.append(self.points[0])
        target_index = 0

        # We loop on the points
        prev = self.points[0]
        src_distance = 0.0
        dst_distance = target_stroke_size
        for curr in self.points[1:]:
            d = self.points_distance( prev, curr )
            if d > 0:
                prev = curr
                src_distance = src_distance+d

                # The new point need to be inserted into the
                # segment [prev, curr]
                while dst_distance < src_distance:
                     x_dir = curr.x - prev.x
                     y_dir = curr.y - prev.y
                     ratio = (src_distance-dst_distance)/d
                     to_x = x_dir * ratio + prev.x
                     to_y = y_dir * ratio + prev.y
                     new_points.append(GesturePoint(to_x, to_y))
                     dst_distance=self.stroke_length(self.points)/float(sample_points)*(len(new_points))

        # If this happens, we are into troubles...
        if not len(new_points) == sample_points:
            raise ValueError("Invalid number of strokes points; got %d while it should be %d" %(len(new_points), sample_points))

        self.points = new_points
        return True

    def center_stroke(self, offset_x,  offset_y):
        '''
        center_stroke(offset_x=float, offset_y=float)
        Centers the stroke by offseting the points
        '''
        for point in self.points:
            point.x -= offset_x
            point.y -= offset_y

class Gesture:
    '''
    A python implementation of a gesture recognition algorithm by Oleg Dopertchouk
    http://www.gamedev.net/reference/articles/article2039.asp

    Implemented by Jeiel Aranal (chemikhazi@gmail.com), released into the public domain
    '''

    # Tolerance for evaluation using the '==' operator
    DEFAULT_TOLERANCE= 0.1
    recognitionMethod = 'dot_product'

    def __init__(self, tolerance=None, recognitionMethod = 'dot_product'):
        '''
        Gesture([tolerance=float])
        Creates a new gesture with an optional matching tolerance value
        '''
        self.strokes = list()
        self.recognitionMethod = recognitionMethod


        if tolerance is None:
            self.tolerance = Gesture.DEFAULT_TOLERANCE
        else:
            self.tolerance = tolerance

    def _scale_gesture(self):
        ''' Scales down the gesture to a unit of 1 '''
        # map() creates a list of min/max coordinates of the strokes
        # in the gesture and min()/max() pulls the lowest/highest value
        min_x = min(map(lambda stroke: stroke.min_x, self.strokes))
        max_x = max(map(lambda stroke: stroke.max_x, self.strokes))
        min_y = min(map(lambda stroke: stroke.min_y, self.strokes))
        max_y = max(map(lambda stroke: stroke.max_y, self.strokes))
        x_len = max_x - min_x
        y_len = max_y - min_y
        scale_factor = max(x_len, y_len)
        if scale_factor <= 0.0:
            return False
        scale_factor = 1.0/scale_factor
        for stroke in self.strokes:
            stroke.scale_stroke(scale_factor)
        return True

    def _center_gesture(self):
        ''' Centers the Gesture,Point of the gesture '''
        total_x = 0.0
        total_y = 0.0
        total_points = 0

        for stroke in self.strokes:
            # adds up all the points inside the stroke
            stroke_y = reduce(lambda total, pt: total + pt.y, stroke.points, 0.0)
            stroke_x = reduce(lambda total, pt: total + pt.x, stroke.points, 0.0)
            total_y += stroke_y
            total_x += stroke_x
            total_points += len(stroke.points)
        if total_points == 0:
            return False
        # Average to get the offset
        total_x /= total_points
        total_y /= total_points
        # Apply the offset to the strokes
        for stroke in self.strokes:
            stroke.center_stroke(total_x, total_y)
        return True

    def add_stroke(self, point_list=None):
        '''
        add_stroke([point_list=list])
        Adds a stroke to the gesture and returns the Stroke instance
        Optional point_list argument is a list of the mouse points for the stroke
        '''
        self.strokes.append(GestureStroke())
        if isinstance(point_list, list) or isinstance(point_list, tuple):
            for point in point_list:
                if isinstance(point, GesturePoint):
                    self.strokes[-1].points.append(point)
                elif isinstance(point, list) or isinstance(point, tuple):
                    if len(point) < 2 or len(point) > 2:
                        raise ValueError("A stroke entry should only have 2 values")
                    self.strokes[-1].add_point(point[0], point[1])
                else:
                    raise TypeError("The point list should either be tuples of x and y or a list of GesturePoint")
        elif point_list is not None:
            raise ValueError("point_list should be a tuple/list")
        return self.strokes[-1]

    def normalize(self, stroke_samples=32):
        ''' Runs the gesture normalization algorithm and calculates the dot product with self '''
        if not self._scale_gesture() or not self._center_gesture():
            self.gesture_product = False
            return False
        for stroke in self.strokes:
            stroke.normalize_stroke(stroke_samples)
        self.gesture_product = self.dot_product(self)

    def get_rigid_rotation(self, dstpts):
        '''
        Extract the rotation to apply to a group of points to minimize the
        distance to a second group of points. The two groups of points are
        assumed to be centered. This is a simple version that just pick
        an angle based on the first point of the gesture.
        '''
        if len(self.strokes) < 1 or len(self.strokes[0].points) < 1:
            return 0
        if len(dstpts.strokes) < 1 or len(dstpts.strokes[0].points) < 1:
            return 0
        target = Vector( [dstpts.strokes[0].points[0].x, dstpts.strokes[0].points[0].y]  )
        source = Vector( [self.strokes[0].points[0].x, self.strokes[0].points[0].y] )
        return source.angle(target)

    def dot_product(self, comparison_gesture):
        ''' Calculates the dot product of the gesture with another gesture '''
        if len(comparison_gesture.strokes) != len(self.strokes):
            return -1
        if getattr(comparison_gesture, 'gesture_product', True) is False or getattr(self, 'gesture_product', True) is False:
            return -1
        dot_product = 0.0
        for stroke_index, (my_stroke, cmp_stroke) in enumerate( zip(self.strokes, comparison_gesture.strokes) ):
            for pt_index, (my_point, cmp_point) in enumerate( zip(my_stroke.points, cmp_stroke.points) ):
                dot_product += my_point.x * cmp_point.x + my_point.y * cmp_point.y
        return dot_product

    def rotate( self, angle ):
        g = Gesture()
        for stroke in self.strokes:
            tmp = []
            for j in stroke.points:
                v = Vector([j.x, j.y]).rotate(angle)
                tmp.append( v )
            g.add_stroke( tmp )
        g.gesture_product = g.dot_product(g)
        return g

    def _get_dot_product_score(self, comparison_gesture, rotation_invariant):
        ''' Returns the matching score of the gesture against another gesture '''
        if isinstance(comparison_gesture, Gesture):
            if rotation_invariant:
                # get orientation
                angle = self.get_rigid_rotation( comparison_gesture )

                # rotate the gesture to be in the same frame.
                comparison_gesture = comparison_gesture.rotate( angle )

            # this is the normal "orientation" code.
            score = self.dot_product(comparison_gesture)
            if score <= 0:
                return score
            score /= math.sqrt(self.gesture_product * comparison_gesture.gesture_product)
            return score


    def _calculate_direction_list(self,comparison_gesture):
        my_strokeList  = []
        cmp_strokeList = []
        totalPoints = 0
        #build list of directions of each step in the gestures
        for stroke_index, (my_stroke, cmp_stroke) in enumerate( zip(self.strokes, comparison_gesture.strokes) ):
            my_lastPoint = None
            cmp_lastPoint = None
            my_directionList = []
            cmp_directionList = []
            for pt_index, (my_point, cmp_point) in enumerate( zip(my_stroke.points, cmp_stroke.points) ):
                totalPoints = totalPoints +1
                if my_lastPoint == None or cmp_lastPoint == None:
                    my_lastPoint = my_point
                    cmp_lastPoint = cmp_point
                else:
                    #print my_point
                    #print my_stroke.max_x,my_stroke.max_y,cmp_stroke.min_x,cmp_stroke.min_y
                    #print my_lastPoint,my_point,self._calculatePointDirection(my_lastPoint, my_point)
                    my_directionList.append(self._calculatePointDirection(my_lastPoint, my_point))
                    cmp_directionList.append(self._calculatePointDirection(cmp_lastPoint, cmp_point))
            my_strokeList.append(my_directionList)
            cmp_strokeList.append(cmp_directionList)
            #print "my" ,self.label, my_directionList
            #print "cmp" , cmp_directionList
        return (my_strokeList, cmp_strokeList,totalPoints)




    def _get_direction_levenshtein_score(self, comparison_gesture, rotation_invariant):
        ''' Returns the distance between the two gestures using Levenshtein distance '''

        if isinstance(comparison_gesture, Gesture):
            if len(comparison_gesture.strokes) != len(self.strokes):
                return -1

            #try 8 angles if rotation_invariant is True
            if rotation_invariant == True:
                cycles = 8
            else:
                cycles = 1

            bestScore = 0.0
            for i in range(cycles):
                totalDistance = 0
                my_strokeList, cmp_strokeList,totalPoints = self._calculate_direction_list(comparison_gesture)

                for direction_index, (my_directionList, cmp_directionList) in enumerate( zip(my_strokeList, cmp_strokeList) ):
                    totalDistance += self._levenshteinDistance(my_directionList,cmp_directionList,i)
                    #compare the stroke using levenshteinDistance

                score = 1.0 - (float(totalDistance)/float(totalPoints))
                #print "the score of all rounds is", score
                if score> bestScore:
                    bestScore = score
            #print "dang dang the score is", score,totalDistance,totalPoints
            return bestScore

    def _get_direction_custom_score(self, comparison_gesture, rotation_invariant):
        ''' Returns the distance between the two gestures using Levenshtein distance '''

        if isinstance(comparison_gesture, Gesture):
            if len(comparison_gesture.strokes) != len(self.strokes):
                return -1
            #print "using _get_direction_custom_score"
            totalDistance = 0

            my_strokeList, cmp_strokeList,totalPoints = self._calculate_direction_list(comparison_gesture)

            for direction_index, (my_directionList, cmp_directionList) in enumerate( zip(my_strokeList, cmp_strokeList) ):
                for i in range(len(my_directionList)):
                    totalDistance += abs(my_directionList[i]-cmp_directionList[i])
                #compare the stroke using levenshteinDistance

            score = 1.0 - (float(totalDistance)/float(totalPoints*7.0)) #times maximum distance between two numbers
            return score



    def _calculatePointDirection(self,fromPoint, toPoint):
        x = toPoint.x -  fromPoint.x
        y = toPoint.y -  fromPoint.y

        angle = math.degrees(math.atan2(x,y))

        if angle < 0:
            angle = 360-abs(angle)

        direction = int(round(angle/45.0))
        if direction == 8: direction = 0

        return direction

    def _calculateNormalizedPointAngle(self,fromPoint, toPoint):
        x = toPoint.x -  fromPoint.x
        y = toPoint.y -  fromPoint.y

        angle = math.degrees(math.atan2(x,y))

        if angle < 0:
            angle = 360-abs(angle)

        direction = angle/360.0


        return direction


    def calculate_directionAngle_list(self):
        strokeList = []
        totalPoints = 0
        #build list of directions of each step in the gestures
        for stroke in self.strokes:
            lastPoint = None
            directionList = []
            for point in stroke.points:
                totalPoints = totalPoints +1
                if lastPoint == None:
                    lastPoint = point
                else:
                    directionList.append(self._calculateNormalizedPointAngle(lastPoint, point))

            strokeList.append(directionList)
        return strokeList


    def _levenshteinDistance(self, my_directionList, cmp_directionList,angleOffset):
        my_length = len(my_directionList)
        cmp_length = len(cmp_directionList)

        matrix = []
        for i in range(0,my_length+1 ):
            temp = []
            for j in range(0,cmp_length+1):
                temp.append(0)
            matrix.append(temp)

        for i in range(0,my_length+1):
            matrix[i][0] = i

        for j in range(0,cmp_length+1):
            matrix[0][j] = j


        for j in range(1,cmp_length+1):
            for i in range(1,my_length+1 ):
                if ((my_directionList[i-1]+angleOffset)%8) == cmp_directionList[j-1]:
                    matrix[i][j] = matrix[i-1][j-1]
                else:
                    matrix[i][j] = min((matrix[i-1][j]+1),(matrix[i][j-1]+1),(matrix[i-1][j-1]+1))

        #for g in range(len(matrix)):
        #    print matrix[g]

        return matrix[my_length][cmp_length]


    def _get_region_levenshtein_score(self, comparison_gesture, rotation_invariant):
        ''' Returns the distance between the two gestures using Levenshtein distance '''

        if isinstance(comparison_gesture, Gesture):
            if len(comparison_gesture.strokes) != len(self.strokes):
                return -1

            totalDistance = 0
            my_strokeList, cmp_strokeList,totalPoints = self._calculate_region_list(comparison_gesture)

            for direction_index, (my_regionList, cmp_regionList) in enumerate( zip(my_strokeList, cmp_strokeList) ):
                totalDistance += self._levenshteinDistance(my_regionList,cmp_regionList,0)
                #compare the strok using levenshteinDistance

            score = 1.0 - (float(totalDistance)/float(totalPoints))

            return score

    def _calculate_region_list(self,comparison_gesture):
        #Use this type of region mapping:
        #|6|7|8|
        #|3|4|5|
        #|0|1|2|
        #Coordinates can be calculated on the normalized
        #my_stroke.max_x,my_stroke.max_y,cmp_stroke.min_x,cmp_stroke.min_y
        my_strokeList  = []
        cmp_strokeList = []
        totalPoints = 0
        #build list of regions the points visit
        for stroke_index, (my_stroke, cmp_stroke) in enumerate( zip(self.strokes, comparison_gesture.strokes) ):
            my_regionList = []
            cmp_regionList = []
            my_distance_x = abs(my_stroke.min_x - my_stroke.max_x)/3.0
            my_distance_y = abs(my_stroke.min_y - my_stroke.max_y)/3.0
            cmp_distance_x = abs(cmp_stroke.min_x - cmp_stroke.max_x)/3.0
            cmp_distance_y = abs(cmp_stroke.min_y - cmp_stroke.max_y)/3.0
            my_region = None
            cmp_region = None

            for pt_index, (my_point, cmp_point) in enumerate( zip(my_stroke.points, cmp_stroke.points) ):
                totalPoints += 1
                if my_point.x <= 0:
                    if my_point.x < (my_stroke.min_x + my_distance_x):
                        my_region = 0
                    else:
                        my_region = 1
                else:
                    if my_point > (my_stroke.max_x - my_distance_x):
                        my_region = 2
                    else:
                        my_region = 1

                if my_point.y <= 0:
                    if my_point.y < (my_stroke.min_y + my_distance_y):
                        my_region += 0
                    else:
                        my_region += 3
                else:
                    if my_point.y > (my_stroke.max_y - my_distance_y):
                        my_region += 6
                    else:
                        my_region += 3

                if cmp_point.x <= 0:
                    if cmp_point.x < (cmp_stroke.min_x + cmp_distance_x):
                        cmp_region = 0
                    else:
                        cmp_region = 1
                else:
                    if cmp_point > (cmp_stroke.max_x - cmp_distance_x):
                        cmp_region = 2
                    else:
                        cmp_region = 1

                if cmp_point.y <= 0:
                    if cmp_point.y < (cmp_stroke.min_y + cmp_distance_y):
                        cmp_region += 0
                    else:
                        cmp_region += 3
                else:
                    if cmp_point.y > (cmp_stroke.max_y - cmp_distance_y):
                        cmp_region += 6
                    else:
                        cmp_region += 3

                #print cmp_region
                if(cmp_region==0 or cmp_region==2 or cmp_region==4 or cmp_region==6 or cmp_region==8):
                    if(len(cmp_regionList)!=0):
                        if(cmp_regionList[-1] != cmp_region):
                            cmp_regionList.append(cmp_region)
                    else:
                        cmp_regionList.append(cmp_region)

                if(my_region==0 or my_region==2 or my_region==4 or my_region==6 or my_region==8):
                    if(len(my_regionList)!=0):
                        if(my_regionList[-1] != my_region):
                            my_regionList.append(my_region)
                    else:
                        my_regionList.append(my_region)

                #print "ding"



            #print cmp_regionList
            my_strokeList.append(my_regionList)
            cmp_strokeList.append(cmp_regionList)
        return (my_strokeList, cmp_strokeList,totalPoints)

    def create2dMapFromStroke(self, g):
        #create an 2D array
        image = []
        image8 = []
        image16 = []
        image1d = []
        for i in xrange(32):
            image.append([])
            for j in xrange(32):
                image[i].append(0)

        for i in xrange(8):
            image8.append([])
            for j in xrange(8):
                image8[i].append(0)

        for i in xrange(16):
            image16.append([])
            for j in xrange(16):
                image16[i].append(0)


        for stroke in g.strokes:
            n_x = 1.0/abs(stroke.min_x - stroke.max_x) #stroke size x to get normalized scale
            n_y = 1.0/abs(stroke.min_y - stroke.max_y) #stroke size y to get normalized scale
            for p in stroke.points:
                #place point in grid
                grid_x = int((p.x + abs(stroke.min_x))*n_x*31)
                grid_y = int((p.y + abs(stroke.min_y))*n_y*31)
                image[grid_y][grid_x] = 1
                #print grid_x,grid_y


        for i in xrange(16):
            for j in xrange(16):
                image16[i][j] = max(image[(i*2)][j*2],image[(i*2)+1][(j*2)+1],image[(i*2)][(j*2)+1],image[(i*2)+1][(j*2)])

        for i in xrange(8):
            for j in xrange(8):
                image8[i][j] = max(image16[(i*2)][j*2],image16[(i*2)+1][(j*2)+1],image16[(i*2)][(j*2)+1],image16[(i*2)+1][(j*2)])


        for i in range(7,-1,-1):
            for j in range(len(image8)):
                image1d.append(image8[i][j])

       # print ""
       # print "data"
       # print image1d
       # for i in range(31,-1,-1):
       #     print "".join(["%s " % el for el in image[i]])
       # for i in range(15,-1,-1):
       #     print "".join(["%s " % el for el in image16[i]])
       # for i in range(7,-1,-1):
       #     print "".join(["%s " % el for el in image8[i]])
       # print "\\data"




        return image1d


    def _get_ann_image_score(self, comparison_gesture, rotation_invariant,ann):
        ''' Returns the distance between the two gestures using ANN '''


        if ann == None:
            print "ANN not found"
            return 0

        if isinstance(comparison_gesture, Gesture):
            if len(comparison_gesture.strokes) != len(self.strokes):
                return -1

            image = self.create2dMapFromStroke(comparison_gesture)

            indexList = ['S','5','N','Circle','Square','Delete','8']
            
            res = ann.test([[image]])
            #print "label er:",self.label
            #score = res[indexList.index(self.label)]
            maxScore = max(res)
            maxScoreIndex = res.index(maxScore)
            maxScoreLabel = indexList[maxScoreIndex]
            #print comparison_gesture.label,image, "->", res
            #print "testing:",self.label,comparison_gesture.label,res,score,indexList.index(self.label)
            #print "maxScore:",maxScore,"maxScoreIndex:",maxScoreIndex,"maxScoreLabel:",maxScoreLabel
            
           
            
            
            return maxScore,maxScoreLabel


    def _get_ann_direction_score(self, comparison_gesture, rotation_invariant,ann):
        ''' Returns the distance between the two gestures using ANN '''

       

        if ann == None:
            print "ANN not found"
            return 0

        if isinstance(comparison_gesture, Gesture):
            if len(comparison_gesture.strokes) != len(self.strokes):
                return -1


            my_strokeList, dirList ,totalPoints = self._calculate_direction_list(comparison_gesture)
            for i in range(len(dirList[0])):
                dirList[0][i] = dirList[0][i]/7.0

            indexList = ['S','5','N','Circle','Square','Delete','8']
            
            res = ann.test([dirList])
            
            maxScore = max(res)
            maxScoreIndex = res.index(maxScore)
            maxScoreLabel = indexList[maxScoreIndex]
            
            return maxScore,maxScoreLabel

    def get_score(self, comparison_gesture, rotation_invariant=True, recognitionMethod = 'dot_product', ann = None):

        if recognitionMethod == 'dot_product':
            return self._get_dot_product_score(comparison_gesture,rotation_invariant)
        elif recognitionMethod == 'direction_levenshtein':
            return self._get_direction_levenshtein_score(comparison_gesture,rotation_invariant)
        elif recognitionMethod == 'direction_custom':
            return  self._get_direction_custom_score(comparison_gesture,rotation_invariant)
        elif recognitionMethod == 'region_levenshtein':
            return  self._get_region_levenshtein_score(comparison_gesture,rotation_invariant)
        elif recognitionMethod == 'annImage':
            return  self._get_ann_image_score(comparison_gesture,rotation_invariant,ann)
        elif recognitionMethod == 'annDirection':
            return  self._get_ann_direction_score(comparison_gesture,rotation_invariant,ann)

        else:
            print "not a known method"
            return None
            ##legg til nye metoder for sammeligning her ##

    def __eq__(self, comparison_gesture):
        ''' Allows easy comparisons between gesture instances '''
        if isinstance(comparison_gesture, Gesture):
            # If the gestures don't have the same number of strokes, its definitely not the same gesture
            score = self.get_score(comparison_gesture)
            if score > (1.0 - self.tolerance) and score < (1.0 + self.tolerance):
                return True
            else:
                return False
        else:
            return NotImplemented

    def __ne__(self, comparison_gesture):
        result = self.__eq__(comparison_gesture)
        if result is NotImplemented:
            return result
        else:
            return not result
    def __lt__(self, comparison_gesture): raise TypeError("Gesture cannot be evaluated with <")
    def __gt__(self, comparison_gesture): raise TypeError("Gesture cannot be evaluated with >")
    def __le__(self, comparison_gesture): raise TypeError("Gesture cannot be evaluated with <=")
    def __ge__(self, comparison_gesture): raise TypeError("Gesture cannot be evaluated with >=")




print "Alert! Using custom gesture class"